﻿package stego.core
{
    import flash.display.DisplayObject;
    import flash.display.MovieClip;
    import flash.display.Sprite;
    import flash.events.Event;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.script.step.AnimateTo;
    import stego.collision.CollisionEvent;
    import stego.collision.CollisionInfo;
    import stego.util.schedule.Schedule;

    /**
     * ...
     * @author Mike Meyer
     */
    public class GameObject extends MovieClip
    {
        public var dx:Number = 0.0;
        public var dy:Number = 0.0;
        public var dr:Number = 0.0;

        /// workaround for Flash rounding to twips
        private var _x:Number = 0.0;
        public override function get x () : Number { return _x; }
        public override function set x (value:Number) : void { _x = value; super.x = value; }

        /// workaround for Flash rounding to twips
        private var _y:Number = 0.0;
        public override function get y () : Number { return _y; }
        public override function set y (value:Number) : void { _y = value; super.y = value; }

        /// so that physics utilities can tell what they need to ignore
        private var mIsPhysicsEnabled:Boolean;
        public function get PhysicsEnabled():Boolean { return mIsPhysicsEnabled; }
        public function set PhysicsEnabled(value:Boolean):void { mIsPhysicsEnabled = value; }

        /// so that physics utilities can tell what they need to ignore
        private var mIsAffectedByGravity:Boolean;
        public function get AffectedByGravity():Boolean { return mIsAffectedByGravity; }
        public function set AffectedByGravity(value:Boolean):void { mIsAffectedByGravity = value; }

        private var _isAlive:Boolean;
        public function get isAlive():Boolean { return _isAlive; }
        public function set isAlive(value:Boolean):void { _isAlive = value; }

        // check to see if a game object is on screen
        public function get isOnScreen():Boolean
        {
            var rect:Rectangle = getBounds(stage);
            return    (rect.left   < stage.stageWidth)
                   && (rect.right  > 0)
                   && (rect.top    < stage.stageHeight)
                   && (rect.bottom > 0) ;
        }

        public var scheduler:Schedule = new Schedule();

        /**
         * GameObjectHolder for managing internal gameobjects
         */
        private var _holder:GameObjectHolder;
        public function get InternalGameObjects():GameObjectHolder { return _holder; }

        protected var mListeners:Array;

        /**
         * Get the coordinates of the center of this game object's collision box
         */
        public function get center():Point
        {
            var bounds:Rectangle = GetCollisionBounds(parent);
            return new Point( 0.5 * (bounds.left + bounds.right),
                              0.5 * (bounds.top + bounds.bottom) );
        }

        /**
         * initial setup and preparation for reuse
         */
        public function Init():void
        {
            this.x = 0.0;
            this.y = 0.0;
            this.dx = 0.0;
            this.dy = 0.0;
            this.visible = true;
            mIsPhysicsEnabled = true;
            mIsAffectedByGravity = true;
            mListeners = new Array();
            _holder = new GameObjectHolder(this);
            isAlive = true;
        }

        public function Integrate(elapsedTime:Number):void
        {
            this.x += dx * elapsedTime;
            this.y += dy * elapsedTime;
            this.rotation += dr * elapsedTime;

            InternalGameObjects.Integrate(elapsedTime);
        }

        /**
         * per-frame updates
         * @param	elapsedTime - how long the last frame was
         */
        public function Update(elapsedTime:Number):void
        {
            InternalGameObjects.Update(elapsedTime);
            scheduler.Update(elapsedTime);
        }

        /**
         * called after all the final positions have been figured
         * just before drawing
         */
        public function PreDraw():void
        {
            InternalGameObjects.PreDraw();
        }

        /**
         * Get the collision bounds for collision testing
         */
        public function GetCollisionBounds(targetCoordinateSpace:DisplayObject = null):Rectangle
        {
            if (null == targetCoordinateSpace)
            {
                targetCoordinateSpace = root;
            }

            return this.getBounds(targetCoordinateSpace);
        }

        /**
         * cleanup for when we are done with this instance for a while.
         * May still get reused!
         */
        public function Shutdown():void
        {
            RemoveAllListeners();
            scheduler.Shutdown();
        }

        /**
         * overloaded so that we can automagically clean up listeners in shutdown
         */
        public override function addEventListener(type:String, listener:Function, useCapture:Boolean = false, priority:int = 0, useWeakReference:Boolean = false):void
        {
            mListeners.push({type:type, listener:listener, useCapture:useCapture});
            super.addEventListener(type, listener, useCapture, priority, useWeakReference);
        }

        /**
         * remove all registered event listeners
         */
        public function RemoveAllListeners():void
        {
            for each (var o:Object in mListeners)
            {
                this.removeEventListener(o.type, o.listener, o.useCapture);
            }
            mListeners = null;
        }

        /**
         * override to deal with collisions
         */
        public function CollisionOccurred(ci:CollisionInfo, object:GameObject):void
        {
            dispatchEvent(new CollisionEvent(ci, object));

            if (!CanPassThrough(ci, object))
            {
                var collisionBox:Rectangle = this.GetCollisionBounds(parent);

                this.x = ci.location.x;
                if ( 0.0 < ci.normal.x )
                {
                    this.x += (0.5 * collisionBox.width);
                }
                else if ( 0.0 > ci.normal.x )
                {
                    this.x -= (0.5 * collisionBox.width);
                }

                this.y = ci.location.y;
                if ( 0.0 < ci.normal.y )
                {
                    this.y += (0.5 * collisionBox.height);
                }
                else if ( 0.0 > ci.normal.y )
                {
                    this.y -= (0.5 * collisionBox.height);
                }

                if (ci.normal.x != 0.0)
                {
                    this.dx = 0.0;
                }
                if (ci.normal.y != 0.0)
                {
                    this.dy = 0.0;
                }
            }
        }

        /**
         * override to let yourself go through things
         */
        public function CanPassThrough(ci:CollisionInfo, object:GameObject):Boolean
        {
            return false;
        }

        /**
         * start an animation
         * @param	objectToAnimate - what we're animating
         * @param	startValues - where it should start at.  example: {x:100}
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function Animate(startValues:Object, endValues:Object, animDuration:Number=1.0):void
        {
            scheduler.RunScript(new stego.script.step.Animate(this, startValues, endValues, animDuration));
        }

        /**
         * start an animation
         * @param	objectToAnimate - what we're animating
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function AnimateTo(endValues:Object, animDuration:Number=1.0):void
        {
            scheduler.RunScript(new stego.script.step.AnimateTo(this, endValues, animDuration));
        }
    }
}
